(* 
  Coq code for the paper:
  Interfacing Hoare Logic and Type Systems for Foundational Proof-Carrying Code
  by Nadeem Abdul Hamid and Zhong Shao

  Coq version 8.0

  This code is a prototype version only -- proofs are mostly complete.
*)

(*
  Coq Version 8.0 *)


(**************************** Type level option ******************************)
Section optT.
Set Implicit Arguments.
Unset Strict Implicit.
Variable A : Type.
Inductive optT : Type :=
  | someT : A -> optT
  | noneT : optT.

Inductive listT : Type :=
  | nilT : listT
  | consT : A -> listT -> listT.

Fixpoint InT (a : A) (l : listT) {struct l} : Prop :=
  match l with
  | nilT => False
  | consT b m => b = a \/ InT a m
  end.

End optT.



(*********************** Custom tactics **************************************)
Ltac IAssumption :=
  match goal with
  | HH:(_ -> ?X1) |- ?X1 => apply HH; eauto
  end.
Ltac IAssumptionWith v :=
  match goal with
  | HH:(_ -> ?X1) |- ?X1 => apply HH with v; eauto
  end.
Ltac IAssumptionWith2 v1 v2 :=
  match goal with
  | HH:(_ -> ?X1) |- ?X1 => apply HH with v1 v2; eauto
  end.
Ltac IAssumptionWith3 v1 v2 v3 :=
  match goal with
  | HH:(_ -> ?X1) |- ?X1 => apply HH with v1 v2 v3; eauto
  end.

Ltac SplitIn hyp := elim hyp; clear hyp; intros.

Ltac SplitAllHyps :=
  repeat
   match goal with
   |  (* splits all conjunctions in hypotheses *)
   H:(_ /\ _) |- _ => inversion_clear H; auto
   end.

Ltac MCRewrite v :=
  match goal with
  | H:(v = _) |- _ => rewrite H; auto
  | H:(_ = v) |- _ => rewrite <- H; auto
  | H:(v = _) |- _ => rewrite H; auto
  | H:(_ = v) |- _ => rewrite <- H; auto
  end.

Ltac MCRewrite2 v1 v2 := MCRewrite v1; MCRewrite v2; auto.

Ltac MCRewriteRev v :=
  match goal with
  | H:(v = _) |- _ => rewrite <- H; auto
  | H:(_ = v) |- _ => rewrite H; auto
  | H:(v = _) |- _ => rewrite <- H; auto
  | H:(_ = v) |- _ => rewrite H; auto
  end.

Ltac MCRewriteIn v t :=
  match goal with
  | H:(v = _),H1:t |- _ => rewrite H in H1; auto
  | H:(v = _),H1:(t _) |- _ => rewrite H in H1; auto
  | H:(v = _),H1:(t _ _) |- _ =>
      rewrite H in H1; auto
  | H:(v = _),H1:(t _ _ _) |- _ =>
      rewrite H in H1; auto
  | H:(v = _),H1:(t _ _ _ _) |- _ =>
      rewrite H in H1; auto
  | H:(v = _),H1:(t _ _ _ _ _) |- _ => rewrite H in H1; auto
  | H:(_ = v),H1:t |- _ => rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _) |- _ => rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _) |- _ =>
      rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _ _) |- _ =>
      rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _ _ _) |- _ =>
      rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _ _ _ _) |- _ =>
      rewrite <- H in H1; auto
  | H:(v = _),H1:t |- _ => rewrite H in H1; auto
  | H:(v = _),H1:(t _) |- _ => rewrite H in H1; auto
  | H:(v = _),H1:(t _ _) |- _ =>
      rewrite H in H1; auto
  | H:(v = _),H1:(t _ _ _) |- _ =>
      rewrite H in H1; auto
  | H:(v = _),H1:(t _ _ _ _) |- _ =>
      rewrite H in H1; auto
  | H:(v = _),H1:(t _ _ _ _ _) |- _ => rewrite H in H1; auto
  | H:(_ = v),H1:t |- _ => rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _) |- _ => rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _) |- _ =>
      rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _ _) |- _ =>
      rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _ _ _) |- _ =>
      rewrite <- H in H1; auto
  | H:(_ = v),H1:(t _ _ _ _ _) |- _ => rewrite <- H in H1; auto
  end.

Ltac MCRewriteIn2 v t1 t2 := MCRewriteIn v t1; MCRewriteIn v t2.

Ltac MCRewriteIn3 v t1 t2 t3 :=
  MCRewriteIn v t1; MCRewriteIn v t2; MCRewriteIn v t3.

Ltac MCRewriteIn4 v t1 t2 t3 t4 :=
  MCRewriteIn v t1; MCRewriteIn v t2; MCRewriteIn v t3; MCRewriteIn v t4.

Ltac MCRewriteIn5 v t1 t2 t3 t4 t5 :=
  MCRewriteIn v t1; MCRewriteIn v t2; MCRewriteIn v t3; MCRewriteIn v t4;
   MCRewriteIn v t5.

Ltac MCRewriteRevIn v t :=
  match goal with
  | H:(v = _),H1:t |- _ => rewrite <- H in H1; auto
  | H:(_ = v),H1:t |- _ => rewrite H in H1; auto
  | H:(v = _),H1:t |- _ => rewrite <- H in H1; auto
  | H:(_ = v),H1:t |- _ => rewrite H in H1; auto
  end.


Ltac MCInv t :=
  match goal with
  | H:t |- _ => inversion H; auto
  | H:(t _) |- _ => inversion H; auto
  | H:(t _ _) |- _ => inversion H; auto
  | H:(t _ _ _) |- _ => inversion H; auto
  | H:(t _ _ _ _) |- _ => inversion H; auto
  | H:(t _ _ _ _ _) |- _ => inversion H; auto
  | H:(t _ _ _ _ _ _) |- _ => inversion H; auto
  end.


Ltac MCInvClear t :=
  match goal with
  | H:t |- _ => inversion H; clear H; auto
  | H:(t _) |- _ => inversion H; clear H; auto
  | H:(t _ _) |- _ => inversion H; clear H; auto
  | H:(t _ _ _) |- _ => inversion H; clear H; auto
  | H:(t _ _ _ _) |- _ => inversion H; clear H; auto
  | H:(t _ _ _ _ _) |- _ => inversion H; clear H; auto
  | H:(t _ _ _ _ _ _) |- _ => inversion H; clear H; auto
  end.

Ltac MCElim t :=
  match goal with
  | H:t |- _ => elim H; auto
  | H:(t _) |- _ => elim H; auto
  | H:(t _ _) |- _ => elim H; auto
  | H:(t _ _ _) |- _ => elim H; auto
  | H:(t _ _ _ _) |- _ => elim H; auto
  | H:(t _ _ _ _ _) |- _ => elim H; auto
  | H:(t _ _ _ _ _ _) |- _ => elim H; auto
  end.

Ltac MCElimClear t :=
  match goal with
  | H:t |- _ => elim H; clear H; auto
  | H:(t _) |- _ => elim H; clear H; auto
  | H:(t _ _) |- _ => elim H; clear H; auto
  | H:(t _ _ _) |- _ => elim H; clear H; auto
  | H:(t _ _ _ _) |- _ => elim H; clear H; auto
  | H:(t _ _ _ _ _) |- _ => elim H; clear H; auto
  | H:(t _ _ _ _ _ _) |- _ => elim H; clear H; auto
  end.


Ltac MCNewInd t :=
  match goal with
  | H:t |- _ => induction H; auto
  | H:(t _) |- _ => induction H; auto
  | H:(t _ _) |- _ => induction H; auto
  | H:(t _ _ _) |- _ => induction H; auto
  | H:(t _ _ _ _) |- _ => induction H; auto
  | H:(t _ _ _ _ _) |- _ => induction H; auto
  | H:(t _ _ _ _ _ _) |- _ => induction H; auto
  end.


Ltac MCClear t :=
  match goal with
  | H2:t |- _ => clear H2
  | H2:(t = _) |- _ => clear t H2
  | H2:(_ = t) |- _ => clear t H2
  | H2:(t = _) |- _ => clear t H2
  | H2:(_ = t) |- _ => clear t H2
  end.

Ltac MCClearAll :=
  repeat
   match goal with
   | H:(?X1 = ?X2) |- _ => try MCClear X1; try MCClear X2
   end.


Ltac MCInjection t :=
  try
   match goal with
   | H1:(t = _),H2:(_ = t) |- _ =>
       rewrite H1 in H2; injection H2; intros
   | H1:(t = _),H2:(t = _) |- _ =>
       rewrite H1 in H2; injection H2; intros
   | H1:(_ = t),H2:(_ = t) |- _ =>
       rewrite <- H1 in H2; injection H2; intros
   | H1:(t = _),H2:(_ = t) |- _ =>
       rewrite H1 in H2; injection H2; intros
   | H1:(t = _),H2:(t = _) |- _ =>
       rewrite H1 in H2; injection H2; intros
   | H1:(_ = t),H2:(_ = t) |- _ =>
       rewrite <- H1 in H2; injection H2; intros
   end;
   try
    match goal with
    | H:(Some t = _) |- _ => injection H; intros
    | H:(_ = Some t) |- _ => injection H; intros
    end; auto.

Ltac MCClear2 t1 t2 := try MCClear t1; try MCClear t2.
Ltac MCClear3 t1 t2 t3 := try MCClear t1; try MCClear t2; try MCClear t3.
Ltac MCClear4 t1 t2 t3 t4 :=
  try MCClear t1; try MCClear t2; try MCClear t3; try MCClear t4.
Ltac MCClear5 t1 t2 t3 t4 t5 :=
  try MCClear t1; try MCClear t2; try MCClear t3; try MCClear t4;
   try MCClear t5.
Ltac MCClear6 t1 t2 t3 t4 t5 t6 :=
  try MCClear t1; try MCClear t2; try MCClear t3; try MCClear t4;
   try MCClear t5; try MCClear t6.
Ltac MCClear7 t1 t2 t3 t4 t5 t6 t7 :=
  try MCClear t1; try MCClear t2; try MCClear t3; try MCClear t4;
   try MCClear t5; try MCClear t6; try MCClear t7. Ltac UnfoldSplit trm hyp := unfold trm in hyp; SplitAllHyps.
Ltac IntrosSplitC := intros; SplitAllHyps; try split.
Ltac IntrosSplitH := intros; SplitAllHyps.
Ltac UnfoldFold trm := unfold trm in |- *; fold trm in |- *. Ltac UnfoldFoldIn trm hyp := unfold trm in hyp; fold trm in hyp; SplitAllHyps.

Ltac MCUnfoldIn v t :=
  match goal with
  | H:t |- _ => unfold v in H; auto
  | H:(t _) |- _ => unfold v in H; auto
  | H:(t _ _) |- _ => unfold v in H; auto
  | H:(t _ _ _) |- _ => unfold v in H; auto
  | H:(t _ _ _ _) |- _ => unfold v in H; auto
  | H:(t _ _ _ _ _) |- _ => unfold v in H; auto
  | H:(t _ _ _ _ _ _) |- _ => unfold v in H; auto
  end.
Ltac MCUnfoldInSelf t := MCUnfoldIn t t.
Ltac MCUnfoldFoldIn v t :=
  match goal with
  | H:t |- _ => UnfoldFoldIn v H; auto
  end.
Ltac MCUnfoldIn2 v1 v2 t :=
  match goal with
  | H:t |- _ => unfold v1, v2 in H; auto
  end.
Ltac MCUnfoldIn3 v1 v2 v3 t :=
  match goal with
  | H:t |- _ => unfold v1, v2, v3 in H; auto
  end.



Ltac ElimEx hyp var := elim hyp; clear hyp; intros var hyp; SplitAllHyps.

Ltac ElimEx2 hyp var1 var2 := ElimEx hyp var1; ElimEx hyp var2.
Ltac ElimEx3 hyp v1 v2 v3 := ElimEx hyp v1; ElimEx hyp v2; ElimEx hyp v3.
Ltac ElimEx4 hyp v1 v2 v3 v4 :=
  ElimEx hyp v1; ElimEx hyp v2; ElimEx hyp v3; ElimEx hyp v4.
Ltac ElimEx5 hyp v1 v2 v3 v4 v5 :=
  ElimEx hyp v1; ElimEx hyp v2; ElimEx hyp v3; ElimEx hyp v4; ElimEx hyp v5.
Ltac ElimEx6 hyp v1 v2 v3 v4 v5 v6 :=
  ElimEx hyp v1; ElimEx hyp v2; ElimEx hyp v3; ElimEx hyp v4; ElimEx hyp v5;
   ElimEx hyp v6.
Ltac ElimEx7 hyp v1 v2 v3 v4 v5 v6 v7 :=
  ElimEx hyp v1; ElimEx hyp v2; ElimEx hyp v3; ElimEx hyp v4; ElimEx hyp v5;
   ElimEx hyp v6; ElimEx hyp v7.

Ltac MCElimEx var t :=
  match goal with
  | H:t |- _ => ElimEx H var
  | H:(t _) |- _ => ElimEx H var
  | H:(t _ _) |- _ => ElimEx H var
  | H:(t _ _ _) |- _ => ElimEx H var
  | H:(t _ _ _ _) |- _ => ElimEx H var
  | H:(t _ _ _ _ _) |- _ => ElimEx H var
  | H:(ex _) |- _ => ElimEx H var
  | H:(ex _) |- _ => ElimEx H var
  end.
Ltac MCElimEx2 var1 var2 t :=
  match goal with
  | H:t |- _ => ElimEx2 H var1 var2
  | H:(ex _) |- _ => ElimEx2 H var1 var2
  | H:(ex _) |- _ => ElimEx2 H var1 var2
  end.
Ltac MCElimEx3 var1 var2 var3 t :=
  match goal with
  | H:t |- _ => ElimEx3 H var1 var2 var3
  | H:(ex _) |- _ => ElimEx3 H var1 var2 var3
  | H:(ex _) |- _ => ElimEx3 H var1 var2 var3
  end.
Ltac MCElimEx4 var1 var2 var3 var4 t :=
  match goal with
  | H:t |- _ => ElimEx4 H var1 var2 var3 var4
  | H:(ex _) |- _ => ElimEx4 H var1 var2 var3 var4
  | H:(ex _) |- _ => ElimEx4 H var1 var2 var3 var4
  end.

Ltac ElimExIn hyp var := ElimEx hyp var; exists var.

Ltac Exists2 a b := exists a; exists b.
Ltac Exists3 a b c := Exists2 a b; exists c.
Ltac Exists4 a b c d := Exists3 a b c; exists d.

Ltac AbsurdEqBase :=
  match goal with
  | _:(?X1 <> ?X1) |- _ => absurd (X1 = X1); auto
  | _:(?X1 <> ?X1) |- _ => absurd (X1 = X1); auto
  end.
Ltac AbsurdEq :=
  match goal with
  | H1:(?X1 = ?X2),H2:(?X1 <> ?X2) |- _ =>
      rewrite H1 in H2; AbsurdEqBase
  | H1:(?X1 = ?X2),H2:(?X2 <> ?X1) |- _ =>
      rewrite H1 in H2; AbsurdEqBase
  | H1:(?X1 = ?X2),H2:(?X1 <> ?X2) |- _ =>
      rewrite H1 in H2; AbsurdEqBase
  | H1:(?X1 = ?X2),H2:(?X2 <> ?X1) |- _ =>
      rewrite H1 in H2; AbsurdEqBase
  | _:(?X1 <> ?X1) |- _ => absurd (X1 = X1); auto
  | _:(?X1 <> ?X1) |- _ =>
      absurd (X1 = X1); auto
  | H1:(?X1 = Some _),H2:(?X1 = None) |- _ =>
      rewrite H1 in H2; inversion H2; auto
  | H1:(?X1 = someT _),H2:(?X1 = noneT _) |- _ =>
      rewrite H1 in H2; inversion H2; auto
  | H:(Some _ = None) |- _ => inversion H; auto
  | H:(None = Some _) |- _ => inversion H; auto
  | H:(noneT _ = someT _) |- _ => inversion H; auto
  | H:(someT _ = noneT _) |- _ => inversion H; auto
  end.

Ltac Absurdity :=
  try match goal with
      | H:False |- _ => absurd False; auto
      end; try AbsurdEq.


(*********************** Custom grammar **************************************)

Infix "**" := prodT (at level 80, right associativity).
Infix ",*" := pairT (at level 80, right associativity).

(*********************** Custom definitions **********************************)


Definition injT1 A P (S : sigT P) : A := match S with
                                         | existT x _ => x
                                         end.

Definition injT2 A P (S : sigT P) : P (injT1 (A:=A) (P:=P) S) :=
  sigT_rect (fun s : sigT P => P (injT1 (A:=A) (P:=P) s))
    (fun (x : A) (p : P x) => p) S.
