 (* 
 * Coq code for the paper
 * 
 *      Interfacing Hoare Logic and Type Systems for FPCC
 *         Nadeem Abdul Hamid, 
 *
 * (For Coq version 8.0)
 *
 *)


Require Export Arith.
Require Import mystuff.


(**************************************************************************)
(* Register definitions and utilities *)
Inductive Reg : Set :=
  | r0 : Reg
  | r1 : Reg
  | r2 : Reg
  | r3 : Reg
  | r4 : Reg
  | r5 : Reg
  | r6 : Reg
  | r7 : Reg.


Definition nat2reg (n : nat) :=
  match n with
  | S O => r0
  | S (S O) => r1
  | S (S (S O)) => r2
  | S (S (S (S O))) => r3
  | S (S (S (S (S O)))) => r4
  | S (S (S (S (S (S O))))) => r5
  | S (S (S (S (S (S (S O)))))) => r6
  | S (S (S (S (S (S (S (S O))))))) => r7
  | _ => r0
  end.


Definition beq_reg : forall r r' : Reg, bool.
intros; assert ({r = r'} + {r <> r'});
 [ intros; compare r r'; auto
 | intros; case H; [ intros; exact true | intros; exact false ] ].
Defined.


Definition reg2nat (r : Reg) :=
  let fix getreg (n : nat) : nat :=
         match n with
         | O => 0
         | S m => if beq_reg (nat2reg n) r then n else getreg m
         end in
  getreg 8.


Lemma beq_reg_neq_false : forall r r', r <> r' -> beq_reg r r' = false.
  intros r r'; case r; case r'; intros; compute in |- *; auto; try AbsurdEq. Qed.
Lemma beq_reg_neq_falseR : forall r r', r' <> r -> beq_reg r r' = false.
  intros; apply beq_reg_neq_false; auto. Qed.
Lemma beq_reg_false_neq : forall r r', r <> r' -> false = beq_reg r r'.
  intros; symmetry  in |- *; apply beq_reg_neq_false; auto. Qed.
Lemma beq_reg_false_neqR : forall r r', r' <> r -> false = beq_reg r r'.
  intros; symmetry  in |- *; apply beq_reg_neq_false; auto. Qed.
Lemma beq_reg_id_true : forall r, beq_reg r r = true.
  intros; induction  r as [| | | | | | | ]; auto.  Qed.
Lemma beq_reg_true_id : forall r, true = beq_reg r r.
  intros; induction  r as [| | | | | | | ]; compute in |- *; auto. Qed.
Lemma beq_reg_eq_true : forall r r', r = r' -> beq_reg r r' = true.
  intros r r'; case r; case r'; intros; compute in |- *; auto;
   try inversion H. Qed.
Lemma beq_reg_eq_trueR : forall r r', r' = r -> beq_reg r r' = true.
  intros; apply beq_reg_eq_true; auto. Qed.
Lemma beq_reg_true_eq : forall r r', r = r' -> true = beq_reg r r'.
  intros; symmetry  in |- *; apply beq_reg_eq_true; auto. Qed.
Lemma beq_reg_true_eqR : forall r r', r' = r -> true = beq_reg r r'.
  intros; symmetry  in |- *; apply beq_reg_eq_true; auto. Qed.

Hint Immediate beq_reg_neq_false beq_reg_neq_falseR beq_reg_false_neq
  beq_reg_false_neqR beq_reg_id_true beq_reg_true_id beq_reg_eq_true
  beq_reg_eq_trueR beq_reg_true_eq beq_reg_true_eqR.

Lemma eqorneq_reg : forall r r' : Reg, r = r' \/ r <> r'.
intros; compare r r'; auto. Qed.

Hint Immediate eqorneq_reg.
