 (* 
 * Coq code for the paper
 * 
 *      Interfacing Hoare Logic and Type Systems for FPCC
 *         Nadeem Abdul Hamid, 
 *
 * (For Coq version 8.0)
 *
 *)


Require Export Arith.
Require Import mystuff.


(**************************************************************************)
(* Register definitions and utilities *)
Inductive Regt : Set :=
  | rA : Regt
  | rB : Regt
  | rC : Regt
  | rD : Regt
  | rE : Regt
  | rF : Regt
  | rG : Regt
  | rH : Regt
  | rI : Regt
  | rJ : Regt
  | rK : Regt
  | rL : Regt
  | rM : Regt
  | rN : Regt
  | rO : Regt
  | rP : Regt.


Definition nat2regt (n : nat) :=
  match n with
  | S O => rA
  | S (S O) => rB
  | S (S (S O)) => rC
  | S (S (S (S O))) => rD
  | S (S (S (S (S O)))) => rE
  | S (S (S (S (S (S O))))) => rF
  | S (S (S (S (S (S (S O)))))) => rG
  | S (S (S (S (S (S (S (S O))))))) => rH
  | S (S (S (S (S (S (S (S (S O)))))))) => rI
  | S (S (S (S (S (S (S (S (S (S O))))))))) => rJ
  | S (S (S (S (S (S (S (S (S (S (S O)))))))))) => rK
  | S (S (S (S (S (S (S (S (S (S (S (S O))))))))))) => rL
  | S (S (S (S (S (S (S (S (S (S (S (S (S O)))))))))))) => rM
  | S (S (S (S (S (S (S (S (S (S (S (S (S (S O))))))))))))) => rN
  | S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O)))))))))))))) => rO
  | S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S O))))))))))))))) => rP
  | _ => rA
  end.


Definition beq_regt : forall r1 r2 : Regt, bool.
intros; assert ({r1 = r2} + {r1 <> r2});
 [ intros; compare r1 r2; auto
 | intros; case H; [ intros; exact true | intros; exact false ] ].
Defined.


Definition regt2nat (r : Regt) :=
  let fix getreg (n : nat) : nat :=
         match n with
         | O => 0
         | S m => if beq_regt (nat2regt n) r then n else getreg m
         end in
  getreg 16.


Lemma beq_regt_neq_false : forall r r', r <> r' -> beq_regt r r' = false.
  intros r r'; case r; case r'; intros; compute in |- *; auto; try AbsurdEq. Qed.
Lemma beq_regt_neq_falseR : forall r r', r' <> r -> beq_regt r r' = false.
  intros; apply beq_regt_neq_false; auto. Qed.
Lemma beq_regt_false_neq : forall r r', r <> r' -> false = beq_regt r r'.
  intros; symmetry  in |- *; apply beq_regt_neq_false; auto. Qed.
Lemma beq_regt_false_neqR : forall r r', r' <> r -> false = beq_regt r r'.
  intros; symmetry  in |- *; apply beq_regt_neq_false; auto. Qed.
Lemma beq_regt_id_true : forall r, beq_regt r r = true.
  intros; induction  r as [| | | | | | | | | | | | | | | ]; auto. Qed.
Lemma beq_regt_true_id : forall r, true = beq_regt r r.
  intros; induction  r as [| | | | | | | | | | | | | | | ]; auto. Qed.
Lemma beq_regt_eq_true : forall r r', r = r' -> beq_regt r r' = true.
  intros r r'; case r; case r'; intros; compute in |- *; auto;
   try inversion H. Qed.
Lemma beq_regt_eq_trueR : forall r r', r' = r -> beq_regt r r' = true.
  intros; apply beq_regt_eq_true; auto. Qed.
Lemma beq_regt_true_eq : forall r r', r = r' -> true = beq_regt r r'.
  intros; symmetry  in |- *; apply beq_regt_eq_true; auto. Qed.
Lemma beq_regt_true_eqR : forall r r', r' = r -> true = beq_regt r r'.
  intros; symmetry  in |- *; apply beq_regt_eq_true; auto. Qed.

Hint Immediate beq_regt_neq_false beq_regt_neq_falseR beq_regt_false_neq
  beq_regt_false_neqR beq_regt_id_true beq_regt_true_id beq_regt_eq_true
  beq_regt_eq_trueR beq_regt_true_eq beq_regt_true_eqR.

Lemma eqorneq_regt : forall r r' : Regt, r = r' \/ r <> r'.
intros; compare r r'; auto. Qed.

Hint Immediate eqorneq_regt.
